/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Interface to Collections
*  SCOPE :              Interface to Collections
*  FILE :               colx.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#pragma implementation "dictx.hxx"
#endif

/* IMPORTS *******************************************************************/

#ifndef __colx_HXX
#include "colx.hxx"
#endif

#ifndef __obj_HXX
#include "obj.hxx"
#endif

#ifndef __dictx_HXX
#include "dictx.hxx"
#endif

#ifndef __dict_HXX
#include "dict.hxx"
#endif

#ifndef __iter_HXX
#include "iter.hxx"
#endif

#ifndef __yooexcept_HXX
#include "yooexcept.hxx"
#endif

/*****************************************************************************/


/*= EXTERNAL FUNCTION =======================================================*/
void LocalColX::domainI(int ind) const
/*=
[DESCRIPTION :]
  Verify domain for insertion
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if ((ind < -1) || (ind >= this->count())) {
    call_throw(YOO_ERR_COL, "List Indice out of bounds") ;
  }  
}


/*= EXTERNAL FUNCTION =======================================================*/
void LocalColX::domainG(int ind) const
/*=
[DESCRIPTION :]
  Verify domain for other manipulations
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if ((ind < 0) || (ind >= this->count())) {
    call_throw(YOO_ERR_COL, "List Indice out of bounds") ;
  }  
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictXGen::init()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (_col != NULL) return ;
  int s = sizeof(Collection) + (YOO_GROWING_STEP * sizeof(ColElem)) ;
  PageHdlMem* pd = yoo_getpage(this) ;
  Cluster* clu = (pd != NULL) ? pd->page()->getCluster() : YOO_TMP_CLU ;
  this->_col = (Collection*)ObjectX::operator new(s, clu, YOO_DICTLEAF_CLID) ;
  new(this->_col) DictLeaf(0) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictXGen::compare(const LocalColX& c,void* (*extract)(const void*),
		   int (*compare)(const void*, const void*)) const 
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (this->count() != c.count()) {
    return 0 ;
  }
  Iterator<ObjectX> iter(c) ;
  for (; iter.more(); iter.next()) {
    if (this->member(iter.get(), extract, compare) == -1) {
      return 0 ;
    }
  }
  return 1 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictXGen::insert(const LocalColX& c,void* (*extract)(const void*),
		   int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Iterator<ObjectX> iter(c) ;
  for (; iter.more(); iter.next()) {
    this->insert(iter.get(), extract, compare) ;
  }  
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictXGen::insert(const yoo_ObjectX& obj,void* (*extract)(const void*),
		   int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->init() ;
  void* key = extract(&obj) ;
  int ind = this->_col->indKey(key, extract, compare) ;
  this->_col = this->_col->insert(&obj, ind) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void DictXGen::insertUniq(const LocalColX& c,void* (*extract)(const void*),
		       int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Iterator<ObjectX> iter(c) ;
  for (; iter.more(); iter.next()) {
    this->insertUniq(iter.get(), extract, compare) ;
  }  
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictXGen::insertUniq(const yoo_ObjectX& o,void* (*extract)(const void*),
		       int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->init() ;
  void *key, *currKey ;
  key = extract(&o) ;
  int ind = this->_col->indKey(key, extract, compare) ;
  if (ind != -1) {
    void* currAtom = this->_col->getElem(ind) ;
    currKey = extract(currAtom) ;
    int flg = compare(currKey, key) ;
    if (flg == 0) {
      return FALSE ;
    }
  }
  this->_col = this->_col->insert(&o, ind) ;
  return TRUE ;
} 


/*= EXTERNAL FUNCTION =======================================================*/
yoo_ObjectX
DictXGen::getFirst(const void* key, void* (*extract)(const void*),
		int (*compare)(const void*, const void*)) const
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (this->count() == 0) {
    return NULL ;
  }
  int ind = this->_col->indKey(key, extract, compare) ;
  if (ind != -1) {
    void* atom = this->_col->getElem(ind) ; 
    void* currKey = extract(atom) ;
    int flg = compare(currKey, key) ;
    if (flg == 0) {
      return *(yoo_ObjectX*)atom ;
    }
  }
  return NULL ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref< ListX<ObjectX> >
DictXGen::getAll(const void* key, void* (*extract)(const void*),
	      int (*compare)(const void*, const void*)) const
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Ref< ListX<ObjectX> > listRes = new(YOO_TMP_CLU) ListX<ObjectX> ;

  if (this->count() == 0) {
    return listRes ;
  }
  int ind = this->indice1(key, extract, compare) ;
  if (ind == -1) {
    return listRes ;
  }
  void* currAtom = this->_col->getElem(ind) ; 
  void* currKey = extract(currAtom) ;
  int flg = compare(currKey, key) ;
  while (flg == 0) {
    (*listRes) += (*(Ref<ObjectX>*)currAtom) ;
    ind++ ;
    if (ind < this->count()) {
      currAtom = this->_col->getElem(ind) ; 
      currKey = extract(currAtom) ;
      flg = compare(currKey, key) ;
    } else {
      break ;
    }
  }

  return listRes ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictXGen::remove(const yoo_ObjectX& o,void* (*extract)(const void*),
		  int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int pos = this->member(o, extract, compare) ;
  if (pos == -1) {
    return -1 ;
  } else {
    this->_col = this->_col->remove(pos) ;
    return 0 ;
  }   
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictXGen::removeAll(const void* key, void* (*extract)(const void*),
		     int (*compare)(const void*, const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int rem = 0 ;
  int flg ;
  
  if (this->count() == 0) {
    return 0 ;
  }

  int ind = this->indice1(key, extract, compare) ;
  if (ind == -1) {
    return rem ;
  }
  void* currAtom = this->_col->getElem(ind) ; 
  void* currKey = extract(currAtom) ;
  flg = compare(currKey, key) ;
  while (flg == 0) {
    rem++ ;
    this->_col = this->_col->remove(ind) ;
    if ( (ind < this->count()) && (this->count() != 0) ) {
      void* currAtom = this->_col->getElem(ind) ; 
      currKey = extract(currAtom) ;
      flg = compare(currKey, key) ;
    } else {
      break ;
    }
  }
  return rem ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictXGen::member(const yoo_ObjectX& obj, void* (*extract)(const void*),
		       int (*compare)(const void*, const void*)) const
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (this->count() == 0) {
    return -1 ;
  }

  void* key = extract((void*)(&obj)) ;
  int ind = this->indice1(key, extract, compare) ;
  if (ind == -1) {
    return -1 ;
  }
  void* currAtom = this->_col->getElem(ind) ; 
  void* currKey = extract(currAtom) ;
  int flg = compare(currKey, key) ;
  while (flg == 0) {
    if (*(yoo_ObjectX*)currAtom == obj) {
      return ind ;
    }
    ind++ ;
    if (ind < this->count()) {
      currAtom = this->_col->getElem(ind) ; 
      currKey = extract(currAtom) ;
      flg = compare(currKey, key) ;
    } else {
      break ;
    }
  }
  return -1 ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
int DictXGen::indice1(const void* key, void* (*extract)(const void*),
			int (*compare)(const void*, const void*)) const
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if (this->count() == 0) {
    return -1 ;
  }
  
  int ind = this->_col->indKey(key, extract, compare) ;
  if (ind <= 0) {
    return ind ;
  }

  ind-- ;
  void* currAtom = this->_col->getElem(ind) ; 
  void* currKey = extract(currAtom) ;
  int flg = compare(currKey, key) ;
  while (flg == 0)  {
    ind-- ;
    if (ind < 0) {
      break ;
    }
    currAtom = this->_col->getElem(ind) ; 
    currKey = extract(currAtom) ;
    flg = compare(currKey, key) ;
  }
  return ind+1 ;  
}


